﻿Imports System.Collections.Concurrent
Imports System.Threading
Imports Syncfusion.Windows.Forms.Tools

Public Class Form1

    Private Sub Form1_Load(sender As Object, e As EventArgs) Handles Me.Load
        Init()
    End Sub

    Private Sub Button1_Click(sender As Object, e As EventArgs) Handles Button1.Click
        Init()
    End Sub

#Region "Init"
    Private Sub Init()
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        InitializeTreeView()

        For i As Integer = 1 To 5
            AddNodeToTree("Test" + i.ToString)
        Next
    End Sub
#End Region

#Region "Member"
    Private mCurrentSourceNode As TreeNodeAdv
    Private mCurrentSelectedNode As TreeNodeAdv
    Private mTreeViewDragHighlightTracker As TreeViewAdvDragHighlightTracker
#End Region

#Region "Properties"
    ''' <summary>
    ''' Wurden Daten geändert
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property Changed As Boolean = False
#End Region

#Region "TreeView"
    ''' <summary>
    ''' Strukturbaum initialisieren
    ''' </summary>
    Private Sub InitializeTreeView()
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        'Tree und Cache leeren
        tvKwd.Nodes.Clear()

        With tvKwd
            .ShowPlusMinus = False
            .Font = New Font("Segoe UI", 10, FontStyle.Regular)
            .ItemHeight = 30
            .ShowDragNodeCue = True
            .AllowDrop = True
        End With

        mTreeViewDragHighlightTracker = New TreeViewAdvDragHighlightTracker(tvKwd)
        AddHandler mTreeViewDragHighlightTracker.QueryAllowedPositionsForNode, AddressOf TreeDragDrop_QueryAllowedPositionsForNode
        AddHandler mTreeViewDragHighlightTracker.QueryDragInsertInfo, AddressOf treeViewDragHighlightTracker_QueryDragInsertInfo
    End Sub

    ''' <summary>
    ''' Knoten hinzufügen
    ''' </summary>
    ''' <param name="name"></param>
    Private Sub AddNodeToTree(name As String)
        Dim newNode As New TreeNodeAdv(name)

        If tvKwd.LastVisibleNode Is Nothing Then
            tvKwd.Nodes.Add(newNode)
        Else
            tvKwd.LastVisibleNode.Nodes.Add(newNode)
        End If

        tvKwd.SelectedNode = tvKwd.Nodes(0)

        tvKwd.ExpandAll()
        tvKwd.Refresh()
    End Sub

    ''' <summary>
    ''' Knoten entfernen
    ''' </summary>
    ''' <param name="node">Elternknoten</param>
    ''' <param name="name">Bezeichnung</param>
    Private Sub RemoveNodeFromTree(node As TreeNodeAdv, name As String)
        If node.Text.Equals(name) Then
            'Knoten löschen

            'Kindknoten an darüber liegenden Knoten verschieben
            If node.HasChildren AndAlso node.Parent IsNot Nothing Then
                node.Nodes(0).Move(node.Parent.Nodes)
            End If

            'Knoten löschen
            node.Remove()
        Else
            For Each childnode As TreeNodeAdv In node.Nodes
                If childnode.HasChildren Then RemoveNodeFromTree(childnode, name)

                If childnode.Text.Equals(name) Then
                    If childnode.HasChildren() Then node.Nodes.Add(childnode.Nodes(0))

                    'Kindknoten an darüber liegenden Knoten verschieben
                    If childnode.HasChildren AndAlso childnode.Parent IsNot Nothing Then
                        childnode.Nodes(0).Move(childnode.Parent.Nodes)
                    End If

                    'Knoten löschen
                    childnode.Remove()

                    Exit For
                End If
            Next
        End If
    End Sub

    ''' <summary>
    ''' Knoten selektiert für Bearbeitung (Doppelklick)
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub tvKwd_DoubleClick(sender As Object, e As EventArgs) Handles tvKwd.DoubleClick, tvKwd.AfterSelect

    End Sub

    ''' <summary>
    ''' Wird beim Begin von DragAndDrop ausgeführt
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub tvKwd_ItemDrag(ByVal sender As Object, ByVal e As ItemDragEventArgs) Handles tvKwd.ItemDrag
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        Dim treeViewAdv As TreeViewAdv = CType(IIf(TypeOf sender Is TreeViewAdv, sender, Nothing), TreeViewAdv)

        'Liste der selektierten Knoten
        Dim nodes() As TreeNodeAdv = TryCast(e.Item, TreeNodeAdv())

        'Es wird der erste Knoten aus der Liste verwendet
        Dim node As TreeNodeAdv = nodes(0)
        Dim result As DragDropEffects = treeViewAdv.DoDragDrop(node, DragDropEffects.Move)
    End Sub

    ''' <summary>
    ''' Wird ausgeführt wenn die Maus über einem potentiellen DragAndDrop-Ziel ist
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub tvKwd_DragOver(ByVal sender As Object, ByVal e As DragEventArgs) Handles tvKwd.DragOver
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        Dim droppable As Boolean = True
        Dim destinationNode As TreeNodeAdv = Nothing
        Dim treeView As TreeViewAdv = CType(IIf(TypeOf sender Is TreeViewAdv, sender, Nothing), TreeViewAdv)
        Dim ptInTree As Point = treeView.PointToClient(New Point(e.X, e.Y))
        mCurrentSourceNode = Nothing

        'Ziel muss ein Knoten sein
        If e.Data.GetDataPresent(GetType(TreeNodeAdv)) Then
            'Quelle und Ziel ermitteln
            destinationNode = treeView.GetNodeAtPoint(ptInTree)
            Dim sourceNode As TreeNodeAdv = CType(e.Data.GetData(GetType(TreeNodeAdv)), TreeNodeAdv)
            mCurrentSourceNode = sourceNode
            droppable = True
        Else
            droppable = False
        End If

        'Anzeigen ob DragAndDrop hier beendet werden darf
        If droppable Then
            e.Effect = DragDropEffects.Move
        Else
            e.Effect = DragDropEffects.None
        End If

        'Highlight setzen
        mTreeViewDragHighlightTracker.SetHighlightNode(destinationNode, ptInTree)
    End Sub

    ''' <summary>
    ''' Wird beim Loslassen der Maustaste während des DragAndDrop ausgeführt
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub tvKwd_DragDrop(ByVal sender As Object, ByVal e As DragEventArgs) Handles tvKwd.DragDrop
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        Dim treeView As TreeViewAdv = TryCast(sender, TreeViewAdv)

        'Quelle und Ziel ermitteln
        Dim sourceNode As TreeNodeAdv = CType(e.Data.GetData(GetType(TreeNodeAdv)), TreeNodeAdv)
        Dim destinationNode As TreeNodeAdv = mTreeViewDragHighlightTracker.HighlightNode

        'Zielposition
        'Dim dropPosition As TreeViewDropPositions = mTreeViewDragHighlightTracker.DropPosition

        'Highlighter berreinigen
        mTreeViewDragHighlightTracker.ClearHighlightNode()

        mCurrentSourceNode = Nothing

        If sourceNode Is destinationNode Then Exit Sub

        If destinationNode IsNot Nothing Then
            If GetDepth(sourceNode) > GetDepth(destinationNode) Then
                'In der Hierarchie nach oben schieben

                'Kinder des Ausgangsknoten an den darüber liegenden Knoten anfügen
                If sourceNode.HasChildren Then sourceNode.Nodes(0).Move(sourceNode.Parent.Nodes)

                'Ausgangsknoten vor den Zielknoten schieben
                sourceNode.Move(destinationNode, NodePositions.Previous)

                destinationNode.Move(sourceNode.Nodes)
            Else
                'In der Hierarchie nach unten schieben

                'Zwei übereinander liegende Knoten tauschen die Plätze
                If destinationNode.Parent Is sourceNode Then
                    'Darunter liegenden Knoten über den Ausgangsknoten schieben
                    destinationNode.Move(sourceNode.Parent.Nodes)

                    'Kinder des darunter liegenden Knoten an den Ausgangsknoten anfügen
                    If destinationNode.HasChildren Then destinationNode.Nodes(0).Move(sourceNode.Nodes)

                    'Ausgangsknoten unter den Zielknoten schieben
                    sourceNode.Move(destinationNode.Nodes)
                Else
                    If sourceNode.HasChildren Then sourceNode.Nodes(0).Move(sourceNode.Parent.Nodes)

                    'Kinder des Zielknoten an den Ausgangsknoten anfügen
                    If destinationNode.HasChildren Then destinationNode.Nodes(0).Move(sourceNode.Nodes)

                    'Ausgangsknoten unter den Zielknoten schieben
                    sourceNode.Move(destinationNode.Nodes)
                End If


            End If
        End If

        'Ausgangsknoten selektieren
        treeView.SelectedNode = sourceNode

        Debug.WriteLine($"Nodes in tree: {treeView.GetNodeCount(True)}")

        _Changed = True
    End Sub

    ''' <summary>
    ''' Tiefe des Knotens in der Hierarchie ermitteln
    ''' </summary>
    ''' <param name="node">Knoten</param>
    ''' <returns>Tiefe</returns>
    Private Function GetDepth(node As TreeNodeAdv)
        Return tvKwd.GetPathFromNode(node).Count(Function(c) c = tvKwd.PathSeparator)
    End Function

    ''' <summary>
    ''' Ermittelt ob die DragAndDrop-Operation fortgeführt wird
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub tvKwd_QueryContinueDrag(ByVal sender As Object, ByVal e As QueryContinueDragEventArgs) Handles tvKwd.QueryContinueDrag
        'Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        'Drag-Operation bei Drücken von Escape abbrechen
        If e.EscapePressed Then e.Action = DragAction.Cancel
    End Sub

    ''' <summary>
    ''' Hebt ein gültiges DragAndDrop-Ziel hervor
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub TreeDragDrop_QueryAllowedPositionsForNode(ByVal sender As Object, ByVal e As QueryAllowedPositionsEventArgs)
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        mTreeViewDragHighlightTracker.EdgeSensitivityOnTop = e.HighlightNode.Bounds.Height \ 4

        'Drag-And-Drop nur auf Knoten zulassen
        e.AllowedPositions = TreeViewDropPositions.OnNode

        'Highlight setzen
        e.ShowSelectionHighlight = CanDrop(mCurrentSourceNode, e.HighlightNode) AndAlso e.NewDropPosition = TreeViewDropPositions.OnNode
    End Sub

    ''' <summary>
    ''' Überprüft ob die DragAndDrop-Operation ausgeführt werden darf
    ''' </summary>
    ''' <param name="sourceNode">Quell-Knoten</param>
    ''' <param name="destinationNode">Ziel-Knoten</param>
    ''' <returns></returns>
    Private Function CanDrop(ByVal sourceNode As TreeNodeAdv, ByVal destinationNode As TreeNodeAdv) As Boolean
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        'Kein DragAndDrop in leere Bereiche oder auf sich selbst
        If sourceNode.TreeView IsNot tvKwd OrElse destinationNode Is Nothing OrElse destinationNode Is sourceNode Then
            Debug.WriteLine("False")
            Return False
        Else
            Debug.WriteLine("True")
            Return True
        End If
    End Function

    ''' <summary>
    ''' Farbe für Hilfsline bei DragAndDrop zwischen einzelnen Knoten
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="args"></param>
    Private Sub treeViewDragHighlightTracker_QueryDragInsertInfo(ByVal sender As Object, ByVal args As QueryDragInsertInfoEventArgs)
        Debug.WriteLine(System.Reflection.MethodBase.GetCurrentMethod.Name)

        args.DragInsertColor = Color.Red
    End Sub

#End Region

#Region "IDisposable"
    ''' <summary>
    ''' Aufräumen
    ''' </summary>
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        Try
            If disposing AndAlso components IsNot Nothing Then
                If mTreeViewDragHighlightTracker IsNot Nothing Then
                    RemoveHandler mTreeViewDragHighlightTracker.QueryAllowedPositionsForNode, AddressOf TreeDragDrop_QueryAllowedPositionsForNode
                    mTreeViewDragHighlightTracker = Nothing
                End If
                If components IsNot Nothing Then
                    components.Dispose()
                End If
            End If
        Finally
            MyBase.Dispose(disposing)
        End Try
    End Sub



#End Region

End Class
